<?php

/**
 * GWDK - Gadget Web Development Kit
 * 
 * Copyright (c) 2004-2005, Michael Wallner, IWORKS
 * All rights reserved.
 * 
 * @package     Gadget
 * @author      Michael Wallner <mike@iworks.at>
 * @version     $Revision: 1.5 $
 * 
 * @copyright   2004-2005 Michael Wallner, IWORKS
 * @license     http://dev.iworks.at/gadget/LICENSE.txt GWDK License
 * 
 * @cvs         $Id: Gadget.php,v 1.5 2005/04/15 13:07:46 mike Exp $
 */

/**
 * Requires PEAR 
 */
require_once 'PEAR.php';

/**
 * Application's INI configuration file
 */
if (!defined('GADGET_CONFIG')) {
    define('GADGET_CONFIG', dirname(__FILE__) .'/gadget-config.ini');
}

/**
 * Application's internal charset
 */
if (!defined('GADGET_CHARSET')) {
    define('GADGET_CHARSET', 'iso-8859-1');
}

/**#@+
 * OnError Constants
 */
define('E_NO',      0);
define('E_LOG',     1);
define('E_NOTIFY',  2);
define('E_DISPLAY', 4);
/**#@-*/

/**#@+
 * Provider Constants
 */
define('GADGET_PROVIDER',           'PROVIDER');
define('PROVIDER_CONFIGURATION',    'configuration');
define('PROVIDER_PREFERENCES',      'preferences');
define('PROVIDER_CURRENTUSER',      'currentuser');
define('PROVIDER_AUTOLOGIN',        'autologin');
define('PROVIDER_RIGHTS',           'rights');
define('PROVIDER_CACHE',            'cache');
/**#@-*/

/**#@+
 * Service Constants
 */
define('GADGET_SERVICE',        'SERVICE');
define('SERVICE_LOG',           'log');
define('SERVICE_RIGHTS',        'rights');
define('SERVICE_AUTH',          'auth');
define('SERVICE_AUTOLOGIN',     'autologin');
define('SERVICE_ERRORLOG',      'errorlog');
define('SERVICE_ERRORNOTIFY',   'errornotify');
define('SERVICE_ERRORDISPLAY',  'errordisplay');
define('SERVICE_SECURITY',      'security');
define('SERVICE_MAILACCOUNT',   'mailaccount');
define('SERVICE_RENDER',        'render');
define('SERVICE_STATS',         'stats');
/**#@-*/

/**
 * PHP4 static class members (sic!)
 * @ignore
 */
$GLOBALS['_Gadget'] = array(
    'qWidgets'  => array(),
    'pWidgets'  => array(),
    'pObjects'  => array(),
    'onError'   => E_DISPLAY,
);

/**
 * @ignore
 */
if (PHP_VERSION < 5) {
    eval('function cl'.'one($obj) { return $obj; }');
}

/**
 * Gadget
 * 
 * The static Gadget class provides all methods you will usually need to perform
 * the most tasks needed to be done while working with the Gadget WDK.
 * 
 * It provides methods for retreiving/manipulating configuration, cookies,
 * user preferences and "the current user".  There are two methods that need to
 * be called at script start, respectively at script end, namely Gadget::start()
 * and Gadget::finish(), which will initialize Gadget's infrastructure and
 * render all instantiated widgets.  See the documentation of the methods 
 * itself for more information.
 * 
 * Example: A basic Gadget powered page:
 * <code>
 * require_once 'Gadget.php';
 * 
 * Gadget::start();
 * 
 * // display login box if
 * // visitor is not logged in
 * if (!Gadget::currentUser()) {
 *      Gadget::newWidget('LoginBox');
 * }
 * 
 * // display intro if user is not 
 * // logged in or it's his preference
 * if (!Gadget::currentUser() or Gadget::getPref('display_intro')) {
 *      Gadget::newWidget(array('template' => 'intro.html'));
 * }
 * 
 * // display home page
 * Gadget::newWidget(array('template' => 'index.html'));
 * 
 * // build a tiny menu
 * // NOTE that the point when you have to build the menu
 * // depends on your chosen layout;
 * $menu = &Gadget::newWidget('Menu');
 * $sect = &$menu->addSection('Navigation');
 * $sect->addLink('Home', '/');
 * $sect->addLink('About', '/about.php');
 * $sect = &$menu->addSection('Downloads', '/downloads/');
 * $sect->addLink('GNU/Linux', '/downloads/linux.php');
 * $sect->addLink('MS Win32', '/downloads/wikndows.php');
 * 
 * Gadget::finish();
 * </code>
 * 
 * @package     Gadget
 * @since       1.0
 * @static
 * @final
 */
class Gadget
{
    // {{{ Configuration
    /**
     * Get Config
     * 
     * Retreive a configuration setting.  In the default setup where
     * Gadget_Provider_Configuration is used, the settings are stored
     * within a plain INI file.  Define the GADGET_CONFIG constant
     * to point to your INI configuration file.
     * 
     * @see Gadget_Provider_Configuration, Gadget::setConf()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $conf Configuration section
     * @param   string  $part Configuration directive
     */
    function getConf($conf = null, $part = null)
    {
        return Gadget::callProviderGet(PROVIDER_CONFIGURATION, array($conf, $part));
    }
    
    /**
     * Set Config
     * 
     * Set a configuration setting.  In the default setup where
     * Gadget_Provider_Configuration is used, this only overrides
     * configuration for the time the current script is executed.
     * 
     * @see Gadget_Provider_Configuration, Gadget::getConf()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $conf Configuration section
     * @param   array   $parts Configuration directives
     */
    function setConf($conf, $parts)
    {
        return Gadget::callProviderSet(PROVIDER_CONFIGURATION, array($conf, $parts));
    }
    // }}}
    
    // {{{ Preferences
    /**
     * Get Preferences
     * 
     * Retreive a preferences item or all preferences if $pref is NULL.
     * If the user has no preferences yet or the user is anonym the
     * default preferences (defined by configuration) are being returned.
     * 
     * @see Gadget_Provider_Preferences, Gadget::setPref()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $pref Preferences item
     */
    function getPref($pref = null)
    {
        return Gadget::callProviderGet(PROVIDER_PREFERENCES, array($pref));
    }
    
    /**
     * Set Preferences
     * 
     * Modify the users preferences.  If the user is anonym the modified
     * preferences will be stored in the session.
     * 
     * @see Gadget_Provider_Preferences, Gadget::getPref()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   array   $pref Preferences items.
     */
    function setPref($pref)
    {
        return Gadget::callProviderSet(PROVIDER_PREFERENCES, array($pref));
    }
    // }}}
    
    // {{{ Current User
    /**
     * Current User
     * 
     * Retreive an attribute of the current user (defaults to users id).
     * If $get === true the users DataObject will be returned.  Returns 0 if
     * the user is anonym.
     * 
     * @see Gadget_Provider_Currentuser, Gadget::setCurrentUser()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   mixed   $get The name of the attribute or true for the whole
     *                  user object.
     */
    function currentUser($get = false)
    {
        return Gadget::callProviderGet(PROVIDER_CURRENTUSER, array($get));
    }
    
    /**
     * Set Current User
     * 
     * Set the current user through its id.
     * 
     * @see Gadget_Provider_Currentuser, Gadget::currentUser()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   int     $id Users id.
     */
    function setCurrentUser($id)
    {
        return Gadget::callProviderSet(PROVIDER_CURRENTUSER, array($id));
    }
    
    /**
     * Get Current User
     * 
     * Retreive the current users DataObject.
     * 
     * @see Gadget::currentUser()
     * 
     * @static
     * @access  public
     * @return  object|false
     */
    function getCurrentUser()
    {
        return Gadget::callProviderGet(PROVIDER_CURRENTUSER, array(true));
    }
    // }}}

    // {{{ Start & Finish
    /**
     * Start
     * 
     * This starts Gadget and is usually called at the very beginning of your
     * page/script.  It does the following in right this order:
     *  o define TIME constant holding the current unix timestamp
     *  o sets a cookie named "first" if not already set
     *  o filter multiple ampersands (&&&) from the query string
     *  o retreives the configured error modes (triggers the configuration 
     *      provider to read in the configuration)
     *  o calls Gadget::onError() to set configured error handling actions
     *      (triggers PEAR::setErrorHandling(PEAR_ERROR_CALLBACK))
     *  o start a PHP session
     *  o runs the security service
     *  o runs the autologin service
     *  o negotiates I18N through HTTP
     *  o querys the caching provider
     *  o instantiates and returns the PageHeader widget unless $header is FALSE
     * 
     * Returns Gadget::header($headerOptions) by default.
     * 
     * <code>
     *  require_once 'Gadget.php';
     *  $header = &Gadget::start();
     *  // or
     *  // Gadget::start();
     *  // $header = &Gadget::header();
     *  $header->addStyleSheet('about');
     *  Gadget::newWidget(array('template' => 'about.html'));
     *  Gadget::finish();
     * </code>
     * 
     * @see Gadget::header(), Gadget::finish(), Gadget_Widget, PageHeader
     * 
     * @static
     * @access  public
     * @return  object  PageHeader
     * @param   bool    $header
     * @param   array   $headerOptions
     */
    function &start($header = true, $headerOptions = array())
    {
        static $init;
        
        if (!$init) {
        
            $init = define('TIME', (int) Gadget::timer());

            isset($_COOKIE['first']) or Gadget::setCookie('first', (int) TIME);
            
            if (isset($_SERVER['QUERY_STRING'])) {
                $_SERVER['QUERY_STRING'] = preg_replace('/&+/', '&', 
                    $_SERVER['QUERY_STRING']);
            } else {
                $_SERVER['QUERY_STRING'] = '';
            }

            // set configured error modes
            $errormode = E_NO;
            foreach (Gadget::getConf('Error') as $mode => $enabled) {
                if ($enabled && defined($ec = 'E_'. strToUpper($mode))) {
                    $errormode |= constant($ec);
                }
            }
            Gadget::onError($errormode);
            
            // handle PEAR errors
            PEAR::setErrorHandling(PEAR_ERROR_CALLBACK, 
                array('Gadget', 'handlePearError'));
            
            // start session
            @session_name(Gadget::getConf('Gadget', 'session_name'));
            @session_start();
        
            // security service
            if (Gadget::getConf('Security', 'enabled')) {
                Gadget::runService(SERVICE_SECURITY);
            }

            // autologin:
            // if user's logged in
            if (Gadget::currentUser()) {
                // check if the user wants autologin enabled
                if (Gadget::getPref('autologin')) {
                    $login = &Gadget::provider(PROVIDER_AUTOLOGIN);
                    // if not already enabled, enable now
                    $login->get() or $login->set(true);
                    unset($login);
                }
            }
            // else fetch credentials for autologin
            elseif ($cr = Gadget::callProviderGet(PROVIDER_AUTOLOGIN)) {
                // if users autologin is corrupt,
                // it will be deleted automatically
                Gadget::runService(SERVICE_AUTOLOGIN, $cr['alias'], $cr['password']);
            }
            
            // caching
            if (Gadget::getConf('Cache', 'enabled')) {
                if ($cacheData = Gadget::callProviderGet(PROVIDER_CACHE)) {
                    header('X-Process-Time: '. Gadget::timer());
                    echo $cacheData;
                    exit(0);
                }
            }
        }
        
        if ($header) {
            $options = is_array($header) ? $header : $headerOptions;
            return Gadget::header($options);
        }
    }
    
    /**
     * Start & Negotiate I18N
     * 
     * @see Gadget::start()
     * 
     * @static
     * @access  public
     * @return  object  PageHeader widget
     * @param   bool    $header
     * @param   array   $headerOptions
     */
    function &startI18N($header = true, $headerOptions = array())
    {
        $header = &Gadget::start($header, $headerOptions);
        
        class_exists('I18Nv2') or require_once 'I18Nv2.php';
        
        // fetch i18n preferences
        $i18n['language'] = Gadget::getPref('language');
        $i18n['charset']  = Gadget::getPref('charset');
        
        // if not set, negotiate i18n settings from browser
        if (!$i18n['language'] || !$i18n['charset']) {
            // fetch supported languages/charsets from config
            $l = explode(',', Gadget::getConf('Gadget', 'languages'));
            $c = explode(',', Gadget::getConf('Gadget', 'charsets'));
            
            if (count($l) > 1 || count($c) > 1) {
                class_exists('I18Nv2_Negotiator') or 
                    require_once 'I18Nv2/Negotiator.php';
                
                // do the negotiate dance
                $n = &new I18Nv2_Negotiator($l[0], $c[0]);
                $i18n['language'] or $i18n['language'] = $n->getLanguageMatch($l);
                $i18n['charset']  or $i18n['charset']  = $n->getCharsetMatch($c);
                unset($n);
            } else {
                $i18n['language'] = $l[0];
                $i18n['charset']  = $c[0];
            }
            // store as preferences
            Gadget::setPref($i18n);
        }
        
        // set locale -> gettext
        I18Nv2::setLocale($i18n['language']);
        // transparently de/encode from/to server preferred charset
        I18Nv2::autoConv($i18n['charset'], GADGET_CHARSET);
        // I18Nv2 doesn't touch the QUERY_STRING
        $_SERVER['QUERY_STRING'] = iconv($i18n['charset'], GADGET_CHARSET, 
        $_SERVER['QUERY_STRING']);
        
        if ($header) {
            $header->setOptions($i18n);
            return $header;
        }
    }

    /**
     * Finish
     * 
     * Gadget::finish() must be called at sript/page end.  It renders all 
     * widgets remaining in the widget queue one after another (unless
     * $widget->show is FALSE).  By default the PageFooter widget will be 
     * instantiated right before the widget queue gets rendered.
     * 
     * @see Gadget::start(), Gadget_Widget, PageFooter
     * 
     * @static
     * @access  public
     * @return  void
     * @param   bool    $footer Whether to instantiate the PageFooter widget.
     * @param   array   $footerOptions Options for the PageFooter widget.
     */
    function finish($footer = true, $footerOptions = array())
    {
        if ($footer) {
            class_exists('PageFooter') or require_once 'Widgets/PageFooter.php';
            new PageFooter(is_array($footer) ? $footer : $footerOptions);
        }
        
        $r = &Gadget::service(SERVICE_RENDER);
        $c = count($GLOBALS['_Gadget']['qWidgets']);
        // render each widget
        for ($i = 0; $i < $c; $i++) {
            $r->run($GLOBALS['_Gadget']['qWidgets'][$i]);
        }
        unset($r, $GLOBALS['_Gadget']['qWidgets']);
        
        $exectime = Gadget::timer();
        headers_sent() or header('X-Process-Time: '. $exectime);
        Gadget::log('EXECTIME', $exectime);
        
        if (Gadget::getConf('Stats', 'enabled')) {
            Gadget::runService(SERVICE_STATS);
        }
        if (Gadget::getConf('Cache', 'enabled')) {
            Gadget::callProviderSet(PROVIDER_CACHE, null);
        }
    }
    // }}}
    
    // {{{ Widgets
    /**
     * New Widget
     * 
     * This will instantiate and return a new widget $widget, configured by
     * $options.  You need to have a directory named "Widgets" in your include
     * path where your custom widgets reside.  If called only with one
     * parameter being an array, a Gadget_Widget will be returned and the 
     * first parameter will be passed through to Gadget_Widget::setOptions().
     * 
     * @see Gadget, Gadget_Widget, Gadget::start(), Gadget::finish()
     * 
     * @static
     * @access  public
     * @return  object
     * @param   mixed
     * @param   array
     */
    function &newWidget($widget, $options = array())
    {
        if (is_array($widget)) {
            $options = $widget;
            $widget  = 'Widget';
        }
        
        if ($widget == 'Widget' || $widget == 'Pager') {
                $incfile = 'Gadget/'. $widget .'.php';
                $widget  = 'Gadget_'. $widget;
                class_exists($widget) or require_once $incfile;
        } else {
                $file = strtr($widget, '_', '/');
                class_exists($widget) or require_once 'Widgets/'.$file.'.php';
        }
        return new $widget($options);
    }
    
    /**
     * Queue Widget
     * 
     * @static
     * @access  public
     * @return  void
     * @param   object Gadget_Widget $widget
     */
    function queueWidget(&$widget, $ID = null)
    {
        static $id = 0;
        
        isset($ID) or $ID = $id++;
        
        $widget->ID = $ID;
        $GLOBALS['_Gadget']['qWidgets'][$ID] = &$widget;
    }
    
    /**
     * Unqueue Widget
     * 
     * @static
     * @access  public
     * @return  void
     * @param   object Gadget_Widget $widget
     */
    function unqueueWidget(&$widget)
    {
        unset($_GLOBALS['_Gadget']['qWidgets'][$widget->ID]);
    }
    
    /**
     * Requeue Widget
     * 
     * @static
     * @access  public
     * @return  void
     * @param   object  Gadget_Widget $widget
     */
    function requeueWidget(&$widget)
    {
        Gadget::unqueue($widget);
        Gadget::queue($widget);
    }
    
    /**
     * Move Widget
     * 
     * @todo    currently it is only possiple to move the queue upwards
     * 
     * @static
     * @access  public
     * @return  bool
     * @param   object  Gadget_Widget $widget
     * @param   object  Gadget_Widget $after
     */
    function moveWidget(&$widget, &$after)
    {
        if ($widget->ID >= $after->ID) {
            return false;
        }
        
        // get ids of all widgets in the queue
        $IDs = array_keys($GLOBALS['_Gadget']['qWidgets']);
        
        // append $widget to the queue if 
        // $after is the last widget in the queue
        if (end($IDs) == $after->ID) {
            Gadget::requeueWidget($widget);
            return true;
        }
        
        // walk through the queue
        while (null !== $ID = array_shift($IDs)) {
            
            // ignore widgets prior to $widget
            if ($ID < $widget->ID) {
                continue;
            }
            
            // we found the target ID of the widget
            // insert $widget at queue postition $ID
            // and exit the loop
            if ($ID > $after->ID) {
                Gadget::queueWidget($widget, $ID);
                break;
            }
            
            // reorder next widget in queue to current position
            Gadget::queueWidget(
                $GLOBALS['_Gadget']['qWidgets'][reset($IDs)], $ID);
        }
        return true;
    }
    // }}}
    
    // {{{ Singletons for Persitent Objects/Widgets and Services/Providers
    /**
     * Create an persistent Object
     * 
     * This is a general purpose singleton, which can instantiate classes
     * named after the PEAR coding standard or from file $file.  Note that
     * objects instantiated through this method are kept alive and may be used 
     * after their first instantiation.  Also note that these classes may only
     * require one construction parameter.
     * 
     * @static
     * @access  public
     * @return  object
     * @param   string
     * @param   mixed
     * @param   string
     */
    function &persistentObject($class, $params = null, $file = null)
    {
        if (isset($GLOBALS['_Gadget']['pObjects'][$class])) {
            $params and $GLOBALS['_Gadget']['pObjects'][$class]->$class($params);
        } else {
            class_exists($class) or
                require_once $file ? $file : strtr($class, '_', '/') .'.php';
            class_exists($class) and
                $GLOBALS['_Gadget']['pObjects'][$class] = &new $class($params);
        }
        return $GLOBALS['_Gadget']['pObjects'][$class];
    }

    /**
     * Create an persistent Widget
     * 
     * This created persitent widgets like the PageHeader.  This is necessary
     * to be able to add style sheets, javascripts, links etc. to the head
     * section within any widget by retreiving the PageHeader widget through
     * &Gadget::header();.
     * 
     * @see Gadget::header(), Gadget::breadcrumbs()
     * 
     * @static
     * @access  public
     * @return  object
     * @param   string
     * @param   mixed
     */
    function &persistentWidget($widget, $options = null)
    {
        if (isset($GLOBALS['_Gadget']['pWidgets'][$widget])) {
            $GLOBALS['_Gadget']['pWidgets'][$widget]->setOptions($options);
        } else {
            $GLOBALS['_Gadget']['pWidgets'][$widget] = 
                &Gadget::newWidget($widget, $options);
        }
        return $GLOBALS['_Gadget']['pWidgets'][$widget];
    }
    
    /**
     * Service
     * 
     * Service singleton.  Returns *the* instance of a certain Gadget_Service
     * object.  It's basically a shorthand for:
     * <code>
     *  require_once 'Gadget/Service.php';
     *  Gadget_Service::singleton($service);
     * </code>
     * Note that it does *NOT* call the run() method of the service.
     * 
     * @see Gadget_Service
     * 
     * @static
     * @access  public
     * @return  object  Gadget_Service
     * @param   string  $service Name of the Service.
     */
    function &service($service)
    {
        class_exists('Gadget_Service') or require_once 'Gadget/Service.php';
        return Gadget_Service::singleton($service);
    }
    
    /**
     * Provider
     * 
     * Provider singleton.  Returns *the* instance of a certain Gadget_Provider
     * object.  It's basically a shorthand for:
     * <code>
     *  require_once 'Gadget/Provider.php';
     *  Gadget_Provider::singleton($provider);
     * </code>
     * 
     * @see Gadget_Provider
     * 
     * @static
     * @access  public
     * @return  object  Gadget_Provider
     * @param   string  $provider Name of the Provider.
     */
    function &provider($provider)
    {
        class_exists('Gadget_Provider') or require_once 'Gadget/Provider.php';
        return Gadget_Provider::singleton($provider);
    }
    
    /**
     * Execute Service
     * 
     * This executes the run() method of a certain Gadget_Service and returns
     * the results of the method call.
     * 
     * @see Gadget::service(), Gadget_Service
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $service The service to run.
     * @param   mixed   $args, ... Arguments to pass through to Service::run()
     */
    function runService($service)
    {
        class_exists('Gadget_Service') or require_once 'Gadget/Service.php';
        $a = array_slice(func_get_args(), 1);
        $s = &Gadget_Service::singleton($service);
        return call_user_func_array(array(&$s, 'run'), $a);
    }
    
    /**
     * Call Provider Set
     * 
     * This executes the set() method of a certain Gadget_Provider and returns
     * the results of the method call. Note that the arguments will be passed
     * to call_user_func_array().
     * 
     * @see Gadget::provider(), Gadget::callProviderGet(), Gadget_Provider
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $provider The provider to instantiate.
     * @param   array   $args Argument to pass through to Gadget_Provider::set()
     */
    function callProviderSet($provider, $args = array())
    {
        class_exists('Gadget_Provider') or require_once 'Gadget/Provider.php';
        $p = &Gadget_Provider::singleton($provider);
        return call_user_func_array(array(&$p, 'set'), $args);
    }
    
    /**
     * Call Provider Get
     * 
     * This executes the get() method of a certain Gadget_Provider and returns
     * the results of the method call.  Note that the arguments will be passed
     * to call_user_func_array().
     * 
     * @see Gadget::provider(), Gadget::callProviderSet(), Gadget_Provider
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $provider The provider to instantiate.
     * @param   array   $args Argument to pass through to Gadget_Provider::get()
     */
    function callProviderGet($provider, $args = array())
    {
        class_exists('Gadget_Provider') or require_once 'Gadget/Provider.php';
        $p = &Gadget_Provider::singleton($provider);
        return call_user_func_array(array(&$p, 'get'), $args);
    }
    // }}}
    
    // {{{ Persistent Widgets
    /**
     * Breadcrumbs
     * 
     * This returns the Breadcrumbs widget instance.
     * 
     * @see Breadcrumbs, Gadget::header(), Gadget::persistentWidget()
     * 
     * @static
     * @access  public
     * @return  object  Breadcrumbs
     * @param   array   $options Options for the Breadcrumb widget.
     */
    function &breadcrumbs($options = array())
    {
        return Gadget::persistentWidget('Breadcrumbs', $options);
    }
    
    /**
     * PageHeader
     * 
     * This returns the PageHeader instance.
     * 
     * @see PageHeader, Gadget::start(), Gadget::breadcrumbs(), 
     *      Gadget::persistentWidget()
     * 
     * @static
     * @access  public
     * @return  object  PageHeader
     * @param   array   $options Options for the PageHeader widget.
     */
    function &header($options = array())
    {
        return Gadget::persistentWidget('PageHeader', $options);
    }
    // }}}

    // {{{ Persistent Objects
    /**
     * Locale
     * 
     * This returns the I18Nv2_Locale instance configured with the 
     * locale/language retreived by Gadget::getPref('language').
     * 
     * @see Gadget::persistenObject(), I18Nv2 (http://pear.php.net/I18Nv2)
     * 
     * @static
     * @access  public
     * @return  object  I18Nv2_Locale
     */
    function &locale()
    {
        return Gadget::persistentObject('I18Nv2_Locale', 
            Gadget::getPref('language'));
    }
    // }}}
    
    // {{{ DataObjects
    /**
     * Setup a DB_DataObject
     * 
     * This calls DB_DataObject::factory() to create a new entity unless called
     * without argument, then it just does the necessary work to properly setup
     * DB_DataObject by retreiving its options through 
     * Gadget::getConf('Db_DataObject').
     * 
     * @see DB_DataObject (http://pear.php.net/DB_DataObject)
     * 
     * @access  public
     * @return  mixed
     * @param   string  $DAO Name of the database table.
     */
    function &DAO($DAO = null)
    {
        static $options_set;
        
        if (!$options_set) {
            $options_set = true;
            if (!defined('DB_DATAOBJECT_NO_OVERLOAD')) {
                define('DB_DATAOBJECT_NO_OVERLOAD', true);
            }
            $DO_options = &PEAR::getStaticProperty('DB_DataObject', 'options');
            $DO_options = Gadget::getConf('DB_DataObject');
            $DB_options = &PEAR::getStaticProperty('DB', 'options');
            $DB_options = Gadget::getConf('DB');
            class_exists('DB_DataObject') or require_once 'DB/DataObject.php';
        }
        
        if (isset($DAO)) {
            return DB_DataObject::factory(ucfirst(strtolower($DAO)));
        }
        return $options_set;
    }
    // }}}
    
    // {{{ Login & Logout
    /**
     * Login
     * 
     * Performs the login task.  It will regenerate the session id, set the
     * "current user" and redirect to $_GET['uri'] if this parameter is
     * present.
     * 
     * @access  public
     * @return  void
     * @param   int     $userId The id of the user to be logged in.
     */
    function login($userId)
    {
        session_regenerate_id();
        Gadget::setCurrentUser((int) $userId);
        isset($_GET['uri']) and Gadget::redirectInternal($_GET['uri']);
    }
    
    /**
     * Logout
     * 
     * Performs the logout task.  It'll regenerate the session id, delete all
     * session variables and finally destroy the session.  It will further
     * disable autologin (delete the autologin cookie) and redirect to the
     * home page.
     * 
     * @access  public
     * @return  void
     */
    function logout($redirect = '/')
    {
        @session_start();
        
        $redirect === true and $redirect = Gadget::getPref('logout_redirect');
        
        // disable autologin
        Gadget::callProviderSet(PROVIDER_AUTOLOGIN, false);
        
        // clear session
        session_destroy();
        session_regenerate_id();
        $_SESSION = array();
        
        // redirect
        Gadget::redirectExternal($redirect);
    }
    // }}}
    
    // {{{ Cookies
    /**
     * Set Cookie
     * 
     * This is a convinience method to set an persistent or a session cookie.  
     * Its lifetime is the date of MAXINT (as unix timestamp) if the cookie is
     * persistent, or to 0 if the cookie is supposed to
     * expire at the end of the session.  The parameters will be urlencoded 
     * just like a query string of an URL.  If the $params argument is not 
     * an array it will be passed through to setcookie().
     * 
     * @see Gadget::getCookie()
     * 
     * @static
     * @access  public
     * @return  bool
     * @param   string  $name Cookie name.
     * @param   array   $params Cookie's contents as associative array.
     * @param   bool    $session Whether Cookie is valid for the session only.
     */
    function setCookie($name, $params = array(), $session = false)
    {
        $value = '';
        if (is_array($params)) {
            foreach ($params as $n => $v) {
                $value .= '&'. urlencode($n) .'='. urlencode($v);
            }
        } else {
            $value = $params;
        }
        return setCookie(   $name, $value, $session ? 0 : 2147483647, '/', 
                            Gadget::getConf('Gadget', 'wwwdomain'));
    }
    
    /**
     * Get Cookie
     * 
     * Retrieves the value of an HTTP cookie.  The value will be urldecoded
     * like a query string of an URL unless $parse is FALSE.
     * 
     * @see Gadget::setCookie()
     * 
     * @static
     * @access  public
     * @return  mixed
     * @param   string  $name Cookie name.
     * @param   bool    $parse Whether to parse the value of the cookie like
     *                  a query string of an URL.
     */
    function getCookie($name, $parse = true)
    {
        if (!isset($_COOKIE[$name])) {
            return null;
        }
        if (!$parse) {
            return $_COOKIE[$name];
        }
        parse_str($_COOKIE[$name], $params);
        return $params;
    }
    
    /**
     * Unset Cookie
     * 
     * Unsets an HTTP cookie.  The contents of the cookie $name will be set
     * to null and the date to expire will be set to 1 (unix timestamp).
     * 
     * @see Gadget::setCookie()
     * 
     * @static
     * @access  public
     * @return  bool
     * @param   string  $name Cookie name.
     */
    function unsetCookie($name)
    {
        unset($_COOKIE[$name]);
        return setCookie($name, null, 1, '/', 
            Gadget::getConf('Gadget', 'wwwdomain'));
    }
    // }}}
    
    // {{{ QueryString Processing
    /**
     * Get Params from QueryString
     * 
     * @static
     * @access  public
     * @return  mixed   Requested value of the query string or specified default.
     * @param   string  $name The name of the GET parameter.
     * @param   string  $type The type the value should be casted to.
     * @param   mixed   $def The default value.
     * @param   bool    $del Whether the param/value pair should be stripped 
     *                  from the query string.
     * @param   mixed   $str Callback for sanitizing strings; Note that arrays 
     *                  and object members will be sanitized, too.
     */
    function getQ($name = null, $type = null, $def = null, $del = false, $str = 'strip_tags')
    {
        if (!isset($name)) {
            return $_SERVER['QUERY_STRING'];
        }
        if (!isset($_GET[$name])) {
            return $def;
        }
        if ($del) {
            Gadget::modQ(array($name => null), true);
        }
        if (!strlen($_GET[$name])) {
            return $def;
        }
        
        switch ($type ? $type{0} : 's')
        {
            case 'b':
                return (bool) $_GET[$name];
            break;
            
            case 'i':
            case 'l':
                return (int) $_GET[$name];
            break;
            
            case 'f':
            case 'd':
                return (double) $_GET[$name];
            break;
            
            case 's':
                if ($str) {
                    return call_user_func($str, (string) $_GET[$name]);
                }
                return (string) $_GET[$name];
            break;
            
            case 'a':
                if ($str) {
                    return array_map($str, (array) $_GET[$name]);
                }
                return (array) $_GET[$name];
            break;
            
            case 'o':
                if ($str) {
                    return (object) array_map($str, (array) $_GET[$name]);
                }
                return (object) $_GET[$name];
            break;
            
            default:
                return $def;
        }
    }
    
    /**
     * Remove Params from QueryString
     * 
     * This operates directly on $_SERVER['QUERY_STRING'] and removes parameters
     * specified by $params.  This is very useful if you ise Apache's 
     * mod_rewrite or similar hooks to prettify URLs.  It returns the
     * modifed query string.
     * 
     * <code>
     *  // assume that the pretty URI /home/user/mike?a=1&b=3
     *  // will be rewritten to /home.php?user=mike&a=1&b=3
     *  Gadget::delQ(array('user'));
     *  // now the query string can be appended to 
     *  // any URL without the "user" parameter
     * </code>
     * 
     * @see Gadget_Widget::$queryString
     * 
     * @static
     * @access  public
     * @return  string  modifed query string
     * @param   array   $params Array of URL parameters to remove.
     * @param   bool    $persistent Whether to write the modified query string
     *                  back to $_SERVER['QUERY_STRING'].
     */
    function delQ($params = array(), $persistent = true)
    {
        $remove = array();
        foreach ((array) $params as $param) {
            $remove[$param] = null;
        }
        return Gadget::modQ($remove, $persistent);
    }
    
    /**
     * Modify QueryString
     * 
     * @static
     * @access  public
     * @return  string  The modifed query string
     * @param   array   $params Array of URL parameters to remove.
     * @param   bool    $persistent Whether to write the modified query string
     *                  back to $_SERVER['QUERY_STRING'].
     */
    function modQ($params = array(), $persistent = true)
    {
        parse_str($_SERVER['QUERY_STRING'], $parsedQueryString);
        foreach ((array) $params as $param => $value) {
            if (isset($value)) {
                $parsedQueryString[$param] = $value;
            } else {
                unset($parsedQueryString[$param]);
            }
        }
        
        $queryString = '';
        foreach ($parsedQueryString as $param => $value) {
            $queryString .= '&'. $param .'='. $value;
        }
        
        if ($persistent) {
            $_SERVER['QUERY_STRING'] = $queryString;
        }
        return $queryString;
    }
    // }}}
    
    /**
     * Location
     * 
     * Get the current location (being the request URI without query string).
     * 
     * @see Gadget_Widget::$location
     * 
     * @static
     * @access  public
     * @return  string
     */
    function location()
    {
        static $location;
        if (!$location) {
            $location = preg_replace('/\?.*$/', '', $_SERVER['REQUEST_URI']);
        }
        return $location;
    }
    
    // {{{ Redirects
    /**
     * Redirect
     * 
     * @static
     * @access  public
     * @return  void
     * @param   string  $uri URI to redirect to.
     * @param   array   $params Parameters to append to the query string.
     * @param   bool    $session Whether to append session name and id.
     */
    function redirect($uri, $params = array(), $session = false)
    {
        Gadget::log('REDIRECT', $uri);
        if ($params and $charset = Gadget::getPref('charset')) {
            class_exists('I18Nv2') or require_once 'I18Nv2.php';
            I18Nv2::recursiveIconv($params, GADGET_CHARSET, $charset);
        }
        class_exists('HTTP_Header') or require_once 'HTTP/Header.php';
        HTTP_Header::redirect($uri, $params, $session);
    }
    
    /**
     * Redirect to Login
     * 
     * Redirects to the configured login page.
     * 
     * @see Gadget_Provider_Configration
     * 
     * @static
     * @access  public
     * @return  void
     */
    function redirectLogin()
    {
        Gadget::redirect(Gadget::getConf('Gadget', 'login'), array('uri' => 
            isset($_GET['uri']) ? $_GET['uri'] : $_SERVER['REQUEST_URI']));
    }
    
    /**
     * Redirect Internal
     * 
     * Issues an internal redirect.  That means that the session will be 
     * preserved and the redirect points directly to the specified URL.
     * 
     * @see Gadget::redirect()
     * 
     * @static
     * @access  public
     * @return  void
     * @param   string  $uri URI to redirect to.
     * @param   array   $params Parameters to append to the query string.
     */
    function redirectInternal($uri = null, $params = array())
    {
        isset($uri) or $uri = Gadget::location();
        Gadget::redirect($uri, $params, !Gadget::getCookie(
            session_name(), false));
    }
    
    /**
     * Redirect External
     * 
     * Issues an external redirect.  That means that the session will be 
     * removed and the redirect points to a configured redirection page.
     * This ensures that eventually existing session information will be
     * stripped from the url and will *not* show up in the referrer log
     * of the foreign server.
     * 
     * @see Gadget::redirect()
     * 
     * @static
     * @access  public
     * @return  void
     * @param   string  $uri URI to redirect to.
     */
    function redirectExternal($uri)
    {
        Gadget::redirect(Gadget::getConf('Gadget', 'redirect'), 
            array('uri' => $uri), false);
    }
    
    /**
     * Redirect Back
     * 
     * Attempts to redirect back to previous page.
     * 
     * @access  public
     * @return  void
     */
    function redirectBack($upperLocation = false)
    {
        if (!$upperLocation) {
            $domain = Gadget::getConf('Gadget', 'wwwdomain');
            if (isset($_SERVER['HTTP_REFERER'])) {
                if (preg_match('/^https?:\/\/'. $domain .'/i', 
                        $_SERVER['HTTP_REFERER'])) {
                    return Gadget::redirectInternal($_SERVER['HTTP_REFERER']);
                } else {
                    Gadget::redirect($_SERVER['HTTP_REFERER']);
                }
            }
        }
        $location = Gadget::location();
        return Gadget::redirectInternal(substr($location, 0, strrpos($location, '/')));
    }
    
    // }}}
    
    // {{{ Error Handling
    /**
     * Set Error Handling
     * 
     * Configure what action to take if a PEAR error occurs.
     * Returns the previous error mode.
     * 
     * <code>
     *  // display errors
     *  Gadget::onError(E_DISPLAY);
     *  // display and log errors
     *  Gadget::onError(E_DISPLAY | E_LOG);
     * </code>
     * 
     * @see Gadget::start(), Gadget::onErrorAdd(), Gadget::onErrorSub(),
     *      Gadget_Service_Errorlog, Gadget_Service_Errornotify
     * 
     * @static
     * @access  public
     * @return  int     Previous error mode (bitmask).
     * @param   int     $mode New error mode (bitmask).
     */
    function onError($mode = null)
    {
        $old = $GLOBALS['_Gadget']['onError'];
        isset($mode) and $GLOBALS['_Gadget']['onError'] = $mode;
        return $old;
    }
    
    /**
     * Add Error Mode
     * 
     * Add error mode to the actual error handling.
     * Returns the previous error mode.
     * 
     * @see Gadget::onError()
     * 
     * @static
     * @access  public
     * @return  int     Previous error mode (bitmask).
     * @param   int     $mode Error mode to add (bitmask).
     */
    function onErrorAdd($mode)
    {
        $old = $GLOBALS['_Gadget']['onError'];
        isset($mode) and $GLOBALS['_Gadget']['onError'] |= $mode;
        return $old;
    }
    
    /**
     * Sub Error Mode
     * 
     * Substract error mode from the actual error handling.
     * Returns the previous error mode.
     * 
     * @see Gadget::onError()
     * 
     * @static
     * @access  public
     * @return  int     Previous error mode (bitmask).
     * @param   int     $mode Error mode to substract (bitmask).
     */
    function onErrorSub($mode)
    {
        $old = $GLOBALS['_Gadget']['onError'];
        isset($mode) and $GLOBALS['_Gadget']['onError'] &=~ $mode;
        return $old;
    }
    
    /**
     * Handle PEAR Error
     * 
     * @internal
     * @static
     * @access  public
     * @return  void
     * @param   object PEAR_Error $e
     */
    function handlePearError(&$e)
    {
        if ($mode = Gadget::onError()) {
            $mode & E_DISPLAY and Gadget::runService(SERVICE_ERRORDISPLAY, $e);
            $mode & E_NOTIFY  and Gadget::runService(SERVICE_ERRORNOTIFY, $e);
            $mode & E_LOG     and Gadget::runService(SERVICE_ERRORLOG, $e);
        }
    }
    // }}}
    
    // {{{ Misc Utils
    /**
     * Timer
     * 
     * Returns the time difference from Gadget::start() to the point of calling.
     * 
     * @access  public
     * @return  float
     */
    function timer($to_least = false)
    {
        static $first;
        static $least;
        
        $prior = $least;
        $least = array_sum(explode(' ', microtime()));;
        
        if (!isset($first)) {
            return $first = $least;
        }
        
        return $to_least ? $least - $prior : $least - $first;
    }
    
    /**
     * Log
     * 
     * @static
     * @access  public
     * @return  void
     * @param   string  $source
     * @param   mixed   $message
     */
    function log($source, $message)
    {
        static $log;
        isset($log) or $log = (array) Gadget::getConf('Log');
        
        if (!empty($log['enabled'])) {
            $source = strToLower($source);
            if (isset($log[$source]) && $log[$source]) {
                Gadget::runService(SERVICE_LOG, $source, $message);
            }
        }
    }
    // }}}
}
?>
