<?php

/**
 * Gadget Web Development Kit
 * 
 * Copyright (c) 2004-2005, Michael Wallner, IWORKS
 * All rights reserved.
 * 
 * @package     Gadget
 * @subpackage  Gadget_Service
 * @author      Michael Wallner <mike@iworks.at>
 * @version     $Revision: 1.3 $
 * 
 * @copyright   2004-2005 Michael Wallner, IWORKS
 * @license     http://dev.iworks.at/gadget/LICENSE.txt GWDK License
 * 
 * @cvs         $Id: Security.php,v 1.3 2005/06/10 07:58:39 mike Exp $
 */

/** 
 * Gadget_Service_Security
 *
 * Gadget's basic security service.  Enable with
 * <pre>
 * [Security]
 * enabled              = 1
 * session_hijacking    = 1
 * </pre>
 * in your config.
 * 
 * @package     Gadget
 * @subpackage  Gadget_Service
 * @since       1.0
 */
class Gadget_Service_Security extends Gadget_Service
{
    var $service = SERVICE_SECURITY;
    var $condition;
    
    /**
     * Execute
     * 
     * @access  protected
     * @return  bool
     */
    function execute()
    {
        if (    !$this->check('http_headers')       ||
                !$this->check('request_methods')    ||
                !$this->check('session_hijacking')  ||
                !$this->check('challenge_cookie')   ||
                !$this->check('no_prefetching')) {
            return false;
        }
        return true;
    }
    
    /**
     * Check
     * 
     * @access  protected
     * @return  bool
     * @param   string  $condition
     */
    function check($condition)
    {
        if ($conf = Gadget::getConf('Security', $condition)) {
            $method = '_'. $condition;
            if (!$this->$method($conf)) {
                $this->condition = $condition;
                return false;
            }
        }
        return true;
    }
    
    /**
     * Attempt: Prevent Session Hijacking
     * 
     * @access  private
     * @return  bool
     */
    function _session_hijacking()
    {
        if (isset($_SESSION['SECURITY']['USER_AGENT'])) {
            return $_SESSION['SECURITY']['USER_AGENT'] 
                == @$_SERVER['HTTP_USER_AGENT'];
        }
        if (isset($_SERVER['HTTP_USER_AGENT'])) {
            $_SESSION['SECURITY']['USER_AGENT'] = $_SERVER['HTTP_USER_AGENT'];
        }
        return true;
    }
    
    /**
     * Attempt: Prevent Interception through a Challenge Cookie
     * 
     * @access  private
     * @return  bool
     */
    function _challenge_cookie()
    {
        $new = md5(uniqid('GADGET', 1));
        $usr = Gadget::getCookie('challenge', false);
        $old = &$_SESSION['SECURITY']['CHALLENGE'];
        
        // OK if first request ($old not set)
        if (!$OK = !isset($old['CURR'])) {
            // OK if challenge codes equal
            $OK = $usr === $old['CURR'];
            
            // user may have opened several pages at once
            // this tends to make the whole thingy useless
            if (!$OK && isset($old['PREV']) && TIME - $old['PREV']['TIME'] < 2) {
                $OK = $usr === $old['PREV']['CODE'];
            }
            
            $old['PREV'] = array('TIME' => TIME, 'CODE' => $old['CURR']);
        }

        Gadget::setCookie('challenge', $new, true);
        $old['CURR'] = $new;

        return $OK;
    }
    
    /**
     * Attempt: Check presence of common HTTP headers
     * 
     * @access  private
     * @return  bool
     */
    function _http_headers()
    {
        if (empty($_SERVER['HTTP_HOST'])) {
            return false;
        }
        if (empty($_SERVER['HTTP_USER_AGENT'])) {
            return false;
        }
        return true;
    }
    
    /**
     * Attempt: Prevent unallowed Request Methods
     * 
     * @access  private
     * @return  bool
     * @param   string  $rm
     */
    function _request_methods($rm)
    {
        if (strlen($rm) < 3) {
            $rm = array('GET', 'POST', 'HEAD');
        } else {
            $rm = array_map('strToUpper', array_map('trim', explode(',', $rm)));
        }
        return in_array(strToUpper($_SERVER['REQUEST_METHOD']), $rm);
    }
    
    /**
     * Attempt: Prevent Prefetching
     * 
     * @access  private
     * @return  bool
     */
    function _no_prefetching()
    {
        if (isset($_SERVER['HTTP_X_MOZ']) && $_SERVER['HTTP_X_MOZ'] == 'prefetch') {
            header('Status: 403 Forbidden');
            exit(); /* don't log out! */
        }
        return true;
    }
    
    /**
     * On Failure
     * 
     * @access  protected
     * @return  bool false
     */
    function onFailure()
    {
        parent::onFailure($this->condition);
        Gadget::logout('/');
        return false;
    }
    
}
?>
